library CreatingConfigurations;

import 'Complex.dart';
import 'Vect3D.dart';
import 'Particle.dart';
import 'utils.dart';
import 'dart:math';
import 'Paths.dart';

Complex createMoleculeFromPath(List<Particle> particles, double approximateDistance, Vect3D position,
                                Vect3D velocity, double particleMass, int numberOfComplexMolecule, List<Vect3D>path,
                                Vect3D c_m)
{
    List<Particle> tmp = new List<Particle>();
    int count = 0;
    double s = 0.0;
    Vect3D c = new Vect3D.fromVector(new Vect3D.fromVector(path.elementAt(0)));
    for (int i = 1; i < path.length; i++)
    {
        Vect3D r2 = new Vect3D.fromVector(path.elementAt(i));
        Vect3D r1 = new Vect3D.fromVector(path.elementAt(i - 1));
        if(c_m == null)
            c = c + r2;
        s += (r2 - r1).abs();
    }
    if(c_m == null)
        c_m = c / path.length.toDouble();
    Vect3D r2 = new Vect3D.fromVector(path.elementAt(path.length - 1));
    Vect3D r1 = new Vect3D.fromVector(path.elementAt(0));
    s += (r2 - r1).abs();
    int n = (s / approximateDistance).toInt();
    approximateDistance = s / n;
    particles.add(new Particle.complex(position + r1 - c_m, velocity, numberOfComplexMolecule, count));
    tmp.add(new Particle.complex(position + r1 - c_m, velocity, numberOfComplexMolecule, count));
    count++;

    s = 0.0;
    for (int i = 1; i < path.length; i++)
    {
        Vect3D r2 = new Vect3D.fromVector(path.elementAt(i));
        Vect3D r1 = new Vect3D.fromVector(path.elementAt(i - 1));
        if(s < approximateDistance)
        {
            s += (r2 - r1).abs();
        }
        else
        {
            particles.add(new Particle.complex(position + r2 - c_m, velocity, numberOfComplexMolecule, count));
            tmp.add(new Particle.complex(position + r2 - c_m, velocity, numberOfComplexMolecule, count));
            count++;
            s = 0.0;
        }
    }

    return new Complex(velocity, new Vect3D.fromNumbers(0.0, 0.0, 0.0), numberOfComplexMolecule, count, tmp,
    position, particleMass);
}


Complex createCircleMolecule(List<Particle> particles, double aproximateDistance, double radius, Vect3D position,
                             Vect3D velocity, double particleMass, int numberOfProppantMolecule)
{
    List<Particle> tmp = new List<Particle>();
    int count = 0;
    double deltaTetta = acos(1 - sqr(aproximateDistance / radius) / 2);
    int n = (2 * PI / deltaTetta).toInt();
    deltaTetta = 2 * PI / n;
    for (double tetta = 0.0; tetta < 2 * PI; tetta += deltaTetta)
    {
        particles.add(new Particle.complex(
            position + new Vect3D.fromNumbers(radius * cos(tetta), radius * sin(tetta), 0.0), velocity,
            numberOfProppantMolecule, count));
        tmp.add(new Particle.complex(
            position + new Vect3D.fromNumbers(radius * cos(tetta), radius * sin(tetta), 0.0), velocity,
            numberOfProppantMolecule, count));
        count++;
    }
    return new Complex(velocity, new Vect3D.fromNumbers(0.0, 0.0, 0.0), numberOfProppantMolecule, count, tmp,
    position, particleMass);
}

int createConfig(List<Particle> mas, double drawRadius, double v0, double approximateDistance, List<Complex> complex,
                  int N_PROPPANT, double particleMass,
                  double width, double height, double a0, List<Vect3D> path, Vect3D c_m, double chance)
{
    height -= drawRadius;
    width -= drawRadius;
    double dx = a0;
    for (double x = drawRadius; x <= width; x += dx)
    {
        mas.add(new Particle(new Vect3D.fromNumbers(x, drawRadius, 0.0), new Vect3D(), TYPE.WALL));
        Particle p = new Particle(new Vect3D.fromNumbers(x, height, 0.0), new Vect3D(), TYPE.WALL);
        p.isStatic = true;
        mas.add(p);
    }
    for (double y = drawRadius; y <= height; y += dx)
    {
        mas.add(new Particle(new Vect3D.fromNumbers(2*drawRadius, y, 0.0), new Vect3D(), TYPE.WALL));
        mas.add(new Particle(new Vect3D.fromNumbers(width-drawRadius, y, 0.0), new Vect3D(), TYPE.WALL));
    }
    if(path == null || path.length == 0)
        path = Paths.Ellipse(2.0 * 10.0, 10.0, approximateDistance);
    double diam = -1.0;
    for(int i = 0; i < path.length; i++)
        for(int j = 0; j < path.length; j++)
        {
            if((path.elementAt(i) - path.elementAt(j)).abs() > diam)
                diam = (path.elementAt(i) - path.elementAt(j)).abs();
        }
    diam += a0;
    int nx = (width ~/ diam);
    int ny = (height ~/ diam);
    Random rand = new Random();
    int count = 0;
    for(num i = 0.5; i < nx - 1; i++)
        for(num j = 0.5; j < ny -1; j++)
        {
            if(rand.nextDouble() < chance)
            {
                complex.add(createMoleculeFromPath(mas, approximateDistance,
                    new Vect3D.fromNumbers(i * diam + a0, j * diam + a0, 0.0), new Vect3D.fromNumbers(0.0, 0.0, 0.0),
                    particleMass, count, path, c_m));
                count++;
            }
        }
    return count;
}
