library Simulation;

import 'dart:html';
import 'Vect3D.dart';
import 'dart:math';
import 'Complex.dart';
import 'CreatingConfigurations.dart';
import 'Particle.dart';
import 'utils.dart';
import 'Box.dart';
import 'Paths.dart';

Element _notes = querySelector("#fps");
num _fpsAverage;

void showFps(num fps)
{
    if (_fpsAverage == null) _fpsAverage = fps;
    _fpsAverage = fps * 0.05 + _fpsAverage * 0.95;
    _notes.text = "${_fpsAverage.round()} fps";
}

class Simulation
{
    CanvasElement simCanvas;
    CanvasElement drawCanvas;
    CanvasRenderingContext2D simContext;
    CanvasRenderingContext2D drawContext;
    num width;
    num height;
    Box box;
    num renderTime;


    InputElement ao, concentrationOfParticles, ok, clear, inputDissipation, inputForce, mode,
        n, animationPeriod, a_b;
    Document doc;
    double drawRadius;
    double rScale13;
    double rScaleShift;

    double a_ellipse = 50.0;
    double fo;
    double mass;
    double To;
    double vo;
    double Bo;
    double betta;
    double dt;
    double v0;
    double vParab;
    double a_cut;
    double a0;
    double curStep = 0.0;
    double s_max;
    double k = 0.7;
    double ka_2;
    double deltaWallTop = 0.0;

    double proppantRadius;
    int N_PROPPANT;
    double concentration = 0.1;


    List<Vect3D> path = new List<Vect3D>();
    List <Particle> particles;
    List <Complex> complex;
    Vect3D externalForce;

    Simulation()
    {
        doc = window.document;
        ao = doc.querySelector("#particle_radius");
        n = doc.querySelector("#n");
        a_b = doc.querySelector("#a_b");
//        concentrationOfParticles = doc.querySelector("#concentrationOfParticles");
        ok = doc.querySelector("#ok");
        clear = doc.querySelector("#clear");
        inputDissipation = doc.querySelector("#inputDissipation");
        inputForce = doc.querySelector("#inputForce");
        mode = doc.querySelector("#checkbox");
        simCanvas = doc.querySelector("#sim");
        drawCanvas = doc.querySelector("#draw");
        animationPeriod = doc.querySelector("#animationPeriod");
        initControlPanel();

        simContext = simCanvas.context2D;
        drawContext = drawCanvas.context2D;
        drawCanvas.onMouseDown.listen(onmousedownDrawCanvas);
        drawCanvas.onMouseUp.listen(onmouseup);

        simCanvas.onMouseDown.listen(onmousedownSimCanvas);
        window.onKeyDown.listen(keyDown);

        Rectangle rect = simCanvas.client;
        width = rect.width;
        height = rect.height;
        box = new Box(new Vect3D(), new Vect3D.fromNumbers(2 * width, 2 * height, 0.0));
        init();
        externalForce = force();
        betta = getBetta();
        requestRedraw();
    }

    Vect3D force()
    {
        return new Vect3D.fromNumbers(0.0, 1.0, 0.0) * inputForce.valueAsNumber / 1000.0;
    }

    double getBetta()
    {
        return inputDissipation.valueAsNumber / 100.0;// * Bo;
    }

    void initControlPanel()
    {
        mode.onChange.listen((_)
        => changeMode());

        ao.onChange.listen((_)
        => init());

        n.onChange.listen((_)
        => changeMode());

        a_b.onChange.listen((_)
        => changeMode());

//        concentrationOfParticles.onChange.listen((_)
//        => init());

        inputDissipation.onChange.listen((_)
        {
            betta = getBetta();
        });
        inputForce.onChange.listen((_)
        {
            externalForce = force();
        });
        ok.onClick.listen((_)
        {
            init();
        });
        clear.onClick.listen((_)
        {
            reset();
        });
    }

    void reset()
    {
        drawBackground(drawContext);
        path = new List<Vect3D>();
    }

    changeMode()
    {
        if (mode.checked)
        {
            reset();
            path = Paths.SuperEllipse(a_ellipse, a_ellipse / (a_b.valueAsNumber / 100.0), n.valueAsNumber / 20, 0.001);
//            mouseMoveListening.cancel();
//            mouseLeaveListening.cancel();

            var rect = drawCanvas.getBoundingClientRect();
            for (int i = 0; i < path.length; i++)
            {
                Vect3D r = path.elementAt(i);
                drawBall(drawContext, r + new Vect3D.fromNumbers(0.5 * (rect.right - rect.left),
                0.5 * (rect.bottom - rect.top), 0.0), rScale13, rScaleShift, 'Red');
            }
        }
    }

    void drawBall(CanvasRenderingContext2D context, Vect3D pos, double rScale13, double rScaleShift, String color)
    {

        var gradient = context.createRadialGradient(pos.x, pos.y,
        rScale13, pos.x - rScaleShift, pos.y + rScaleShift, 0);
        gradient.addColorStop(0, "#000000");
        gradient.addColorStop(1, color);
        context.fillStyle = gradient;

        context.beginPath();
        context.arc(pos.x, pos.y, drawRadius, 0, 2 * PI, false);
        context.closePath();
        context.fill();
    }

    void draw([_])
    {
        while(curStep++ < s_max)
        {
            calculate();
            if(curStep % animationPeriod.valueAsNumber == 0)
                break;
        }
        num time = new DateTime.now().millisecondsSinceEpoch;
//        if (renderTime != null) showFps(1000 / (time - renderTime));
        renderTime = time;
        drawBackground(simContext);
        drawParticles(simContext);
        requestRedraw();
    }

    void drawParticles(CanvasRenderingContext2D context)
    {
        for (int i = 0; i < particles.length; i++)
        {
            Particle p = particles.elementAt(i);
            String color;
            switch (p.type)
            {
                case(TYPE.WALL):
                    color = 'Yellow';
                    break;
                case(TYPE.SIMPLE):
                    color = "Blue";
                    break;
                case(TYPE.COMPLEX):
                    color = 'Red';
                    break;
                default:
                    color = "#f3a67f";
                    break;
            }
            drawBall(context, p.r, rScale13, rScaleShift, color);
        }
    }

    void drawBackground(CanvasRenderingContext2D context)
    {
        context.clearRect(0, 0, width, height);
    }

    void requestRedraw()
    {
        window.requestAnimationFrame(draw);
    }

    void calculate()
    {
        double aa = ao.valueAsNumber * ao.valueAsNumber;
        final double b = pow(13.0 / 7.0, 1.0 / 6.0) * ao.valueAsNumber, bb = b * b;
        final double c = 1.0;
        double aa_cut = a_cut * a_cut;
        double f1 = fo * dt / ao.valueAsNumber;
        for (int i = 0; i < particles.length; i++)
        {
            Particle p_i = particles.elementAt(i);
//            if (!box.contatin(p_i.r))
//            {
//                particles.remove(p_i);
//                continue;
//            }
            for (int j = i + 1; j < particles.length; j++)
            {
                Particle p_j = particles.elementAt(j);
                if (p_i.type == TYPE.WALL && p_j.type == TYPE.WALL)
                    continue;
                if (p_i.type == TYPE.COMPLEX && p_j.type == TYPE.COMPLEX &&
                p_i.numberOfProppantMolecule == p_j.numberOfProppantMolecule)
                    continue;
                Vect3D dr = p_i.r - p_j.r;
                if (dr.x.abs() < a_cut && dr.y.abs() < a_cut && dr.z.abs() < a_cut)
                {
                    double dd = dr.len2();
                    if (dd < aa_cut)
                    {
                        if (dd < ka_2)
                            dd = ka_2;
                        double ss = aa / dd;
                        double s4 = ss * ss;
                        double F = (1 - ss * s4) * s4 * s4;
                        if (dd > bb)
                            F *= c * sqr(1 - c * sqr((dd - bb) / (aa_cut - bb)));
                        Vect3D force = dr * (-(f1 * F));
                        p_i.v = p_i.v + force;

                        if (p_i.type == TYPE.COMPLEX)
                        {
                            Complex p = complex.elementAt(p_i.numberOfProppantMolecule);
                            p.w = p.w + (p_i.r - p.pos).vecMul(force) / p.I;
                            p.a = p.a + force;
                        }
                        force = force * (-1.0);
                        p_j.v = p_j.v + force;
                        if (p_j.type == TYPE.COMPLEX)
                        {
                            Complex p = complex.elementAt(p_j.numberOfProppantMolecule);
                            p.w = p.w + (p_j.r - p.pos).vecMul(force) / p.I;
                            p.a = p.a + force;
                        }
                    }
                }

            }
        }

        int curProppant = -1;
        for (int i = 0; i < particles.length; i++)
        {
            Particle p_i = particles.elementAt(i);
            if (p_i.type == TYPE.WALL)
            {
                if(!p_i.isStatic)
                    p_i.r.y += deltaWallTop;
                continue;
            }
            if (p_i.type == TYPE.COMPLEX)
            {
                if (p_i.numberOfProppantMolecule != curProppant)
                {
                    curProppant = p_i.numberOfProppantMolecule;
                    Complex p = complex.elementAt(curProppant);
                    p.v = p.v + externalForce + p.a / p.nParticlesInMolecule.toDouble() - p.v * betta / p.nParticlesInMolecule.toDouble();
                    p.a.reset();
                    p.pos = p.pos + p.v * dt;
                }
                Complex p = complex.elementAt(curProppant);
                Vect3D w = new Vect3D.fromVector(p.w);
                p.w.reset();
                p.angle += w.z * dt;
                p_i.r = p.pos +
                p.particles.elementAt(p_i.numberOfParticleInProppantMolecule).r.rotateZ(p.angle);
            }
            else
            {
                p_i.r = p_i.r + (p_i.v - p_i.v * betta + externalForce) * dt;
            }
        }
        deltaWallTop = 0.0;
    }

    var mouseMoveListening;
    var mouseLeaveListening;
    double x, y;

    void onMouseMove(MouseEvent event)
    {
        Point m = mouseCoordsDraw(event);
        Point cur = event.client;
        drawBall(drawContext, new Vect3D.fromNumbers(m.x, m.y, 0.0), rScale13, rScaleShift, 'Red');
        path.add(new Vect3D.fromNumbers(m.x, m.y, 0.0));
        x = m.x;
        y = m.y;
    }

    Point mouseCoordsDraw(MouseEvent e)
    {
        var rect = drawCanvas.getBoundingClientRect();
        double x = e.client.x - rect.left;
        double y = e.client.y - rect.top;
        return new Point(x, y);
    }

    Point mouseCoordsSim(MouseEvent e)
    {
        var rect = simCanvas.getBoundingClientRect();
        double x = e.client.x - rect.left;
        double y = e.client.y - rect.top;
        return new Point(x, y);
    }

    onmouseup(MouseEvent e)
    {
        if (drawing)
        {
            mouseMoveListening.cancel();
            mouseLeaveListening.cancel();
        }
    }

    onmouseleave(MouseEvent e)
    {
        if (drawing)
        {
            mouseMoveListening.cancel();
            mouseLeaveListening.cancel();
        }
    }

    bool drawing = false;
    Vect3D centerMass = null;

    onmousedownDrawCanvas(MouseEvent e)
    {
        Point m = mouseCoordsDraw(e);
        x = m.x;
        y = m.y;
        if (e.button == 0)
        {
            if (!mode.checked)
            {
                if (!drawing)
                {
                    centerMass = null;
                    path = new List<Vect3D>();
                    drawing = true;
                }
                mouseMoveListening = drawCanvas.onMouseMove.listen(onMouseMove);
                mouseLeaveListening = drawCanvas.onMouseLeave.listen(onmouseleave);
            }
        }
        else
        {
            centerMass = new Vect3D.fromNumbers(x, y, 0.0);
            drawBall(drawContext, new Vect3D.fromNumbers(m.x, m.y, 0.0), rScale13, rScaleShift, 'Yellow');
        }
    }

    onmousedownSimCanvas(MouseEvent e)
    {
        Point m = mouseCoordsSim(e);
        x = m.x;
        y = m.y;
        if (e.button == 0)
            particles.add(new Particle(new Vect3D.fromNumbers(m.x, m.y, 0.0), new Vect3D.rand(10.0),
            TYPE.SIMPLE));
        if (e.button == 1 && path != null && path.length != 0)
            complex.add(createMoleculeFromPath(particles, ao.valueAsNumber,
            new Vect3D.fromNumbers(m.x, m.y, 0.0), new Vect3D.rand(10.0),
            mass, countComplex++, path, centerMass));
    }

    void init()
    {
        drawBackground(drawContext);
        drawing = false;
        drawRadius = ao.valueAsNumber.toDouble() / 2;
        rScale13 = drawRadius * 1.3;
        rScaleShift = drawRadius / 8;
        fo = 1.0;
        mass = 1.0;
        To = 2 * PI * sqrt(ao.valueAsNumber / 6 / fo);
        vo = sqrt(6 * fo * ao.valueAsNumber);
        Bo = 2 * sqrt(6 * fo / ao.valueAsNumber);
        dt = To / 40;
        v0 = 0.05 * vo;
        a_cut = 1.7 * ao.valueAsNumber;
        ka_2 = sqr(k * ao.valueAsNumber);
        s_max = 1e15;
        deltaWallTop = 0.0;

        proppantRadius = 2 * ao.valueAsNumber;
        N_PROPPANT = 10;

        concentration = 1.0;//concentrationOfParticles.valueAsNumber / 1000;
        drawRadius = ao.valueAsNumber / 2;
        particles = new List<Particle>();
        complex = new List<Complex>();
//        if (path != null && path.length != 0)
//            countComplex = createConfig(particles, drawRadius, v0, distance, complex, N_PROPPANT, mass,
//            width, height, ao.valueAsNumber, path, centerMass, concentration);
        if (path != null && path.length != 0)
            countComplex = createConfig(particles, drawRadius, v0, ao.valueAsNumber, complex, N_PROPPANT, mass,
            width, height, ao.valueAsNumber, path, centerMass, concentration);
    }

    int countComplex = 0;

    void keyDown(KeyboardEvent e)
    {
        if (e.keyCode == KeyCode.DOWN)// && deltaWallTop < height - 4 * drawRadius)
        {
            deltaWallTop += ao.valueAsNumber / 25;
        }
        else
        if (e.keyCode == KeyCode.UP)// && deltaWallTop > drawRadius)
        {
            deltaWallTop -= ao.valueAsNumber / 25;
        }
    }
}